print("import: 'bpdb'")
import bpdb

print("import: 'bpython'")
import bpython

print("import: 'bpython.curtsiesfrontend'")
import bpython.curtsiesfrontend

print("import: 'bpython.test'")
import bpython.test

print("import: 'bpython.test.fodder'")
import bpython.test.fodder

print("import: 'bpython.translations'")
import bpython.translations

